/* *****************************************************************
*								   *
*	SB-Prolog						   *
*	Copyright SUNY at Stony Brook, 1986			   *
*								   *
****************************************************************** */

/*-----------------------------------------------------------------
SB-Prolog is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the SB-Prolog General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
SB-Prolog, but only under the conditions described in the
SB-Prolog General Public License.   A copy of this license is
supposed to have been given to you along with SB-Prolog so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies. 
------------------------------------------------------------------ */


/* ---------------------------- $transform1.P ---------------------------- */

/*  "$transform" transforms definitions containing ';' and "not" to straight-
     line code suitable for mode inference.  				   */

/* **********************************************************************
$transform1_export([$transform/4]).

$transform1_use($blist,[$append/3,$member/2,$member1/2]).
$transform1_use($name,[$name/2,_]).
$transform1_use($aux1,[_,_,_,_,_,_,_,_,$logical_or/3]).
$transform1_use($listutil1,[_,$merge/3,_,_,_,_,_,_]).
$transform1_use($inline1,[$inline/2]).
$transform1_use($meta,[$functor/3,$univ/2,_]).

********************************************************************** */


$transform(PredDefs,TPredDefs0,TPredDefs1,AffList) :-
	$trans_get_deps(PredDefs,DepsList),
	$trans_get_affect_list(DepsList,DepsList,AffList),
	$transform0(PredDefs,AffList,TPredDefs0,TPredDefs1).

$transform0([],_,[],[]).
$transform0([pred(P,N,C0,C1,Clauses)|PRest],[pred(P,N,Deps,StL,_)|ARest],
	    [pred(P,N,C0,C1,NClauses0)|NRest0],[pred(P,N,C0,C1,NClauses1)|NRest1]) :-
	((StL =:= 1, $transform0a(Clauses,NClauses0,NClauses1)) ;
	 (StL =\= 1, NClauses0 = Clauses, NClauses1 = Clauses)
	),
	$transform0(PRest,ARest,NRest0,NRest1).

$transform0a([],[],[]).
$transform0a([Cl|ClRest],XCl_a,XCl_b) :-
	$transform1(Cl,XCl00,XCl01),
	$append(XCl00,XCl10,XCl_a),
	$append(XCl01,XCl11,XCl_b),
	$transform0a(ClRest,XCl10,XCl11).

$transform1(Fact,FL,FL) :- Fact = fact(_,_), FL = [Fact].
$transform1(Rule,NewRules0,NewRules1) :-
	Rule = rule(Head,Body,C0,C1),
	($trans_check2(Body) ->
		($transform2(Body,[true],_,[true],NewBodies0,[true],_,[true],NewBodies1),
		 $transform3(Head,C0,C1,NewBodies0,NewRules0),
		 $transform3(Head,C0,C1,NewBodies1,NewRules1)
		) ;
		(NewRules0 = [Rule], NewRules1 = [Rule])
	).

$transform2(','(C1,C2),PathsA,NPathsA,BodiesA,NBodiesA,PathsB,NPathsB,BodiesB,NBodiesB) :-
	$transform2(C2,PathsA,NPaths0A,BodiesA,NBodies0A,PathsB,NPaths0B,BodiesB,NBodies0B),
	$transform2(C1,NPaths0A,NPathsA,NBodies0A,NBodiesA,NPaths0B,NPathsB,NBodies0B,NBodiesB).
$transform2(';'('->'(C0,C1),C2),PathsA,NPathsA,BodiesA,NBodiesA,PathsB,NPathsB,BodiesB,NBodiesB) :-
	!,
	$transform2(';'((C0,!,C1), C2),PathsA,NPathsA,BodiesA,NBodiesA,PathsB,NPathsB,BodiesB,NBodiesB).
$transform2(';'(C1,C2),PathsA,NPathsA,BodiesA,NBodiesA,PathsB,NPathsB,BodiesB,NBodiesB) :-
	$transform2(C1,PathsA,NPaths0A,BodiesA,NBodies0A,PathsB,NPaths0B,BodiesB,NBodies0B),
	$append(NPaths0A,NPaths1A,NPathsA),
	$append(NBodies0A,NBodies1A,NBodiesA),
	$append(NPaths0B,NPaths1B,NPathsB),
	$append(NBodies0B,NBodies1B,NBodiesB),
	$transform2(C2,PathsA,NPaths1A,BodiesA,NBodies1A,PathsB,NPaths1B,BodiesB,NBodies1B).
$transform2('->'(C1,C2),PathsA,NPathsA,BodiesA,NBodiesA,PathsB,NPathsB,BodiesB,NBodiesB) :-
	$transform2(C2,PathsA,NPaths0A,BodiesA,NBodies0A,PathsB,NPaths0B,BodiesB,NBodies0B),
	$transform2(C1,NPaths0A,NPathsA,NBodies0A,NBodiesA,[true],NPaths1B,[true],NBodies1B),
	$transform_arrow(NPaths1B,NPaths0B,NPathsB),
	$transform_arrow(NBodies1B,NBodies0B,NBodiesB).
$transform2(not(C),PathsA,NPathsA,BodiesA,NBodiesA,PathsB,PathsB,BodiesB,NBodiesB) :-
	$append(BodiesA,NBodies0A,NBodiesA),
	$trans_app_all(not(C),BodiesB,NBodiesB),
	$transform2((C,!),[fail],_,[fail],NBodies0A,[fail],_,[fail],_).
					  /* "fail" -> see xhead/13 */
$transform2(C,PathsA,NPathsA,BodiesA,NBodiesA,PathsB,NPathsB,BodiesB,NBodiesB) :-
	$trans_app_all(C,PathsA,NPathsA),
	$trans_app_all(C,BodiesA,NBodiesA),
	$trans_app_all(C,PathsB,NPathsB),
	$trans_app_all(C,BodiesB,NBodiesB).

$trans_app_all(_,[],[]).
$trans_app_all(C,[G1|GRest],[','(C,G1)|GCRest]) :-
	$trans_app_all(C,GRest,GCRest).

$transform_arrow([],_,[]).
$transform_arrow([P|PRest],L,NewP) :-
	$transform_arrow1(L,P,NewP0),
	$append(NewP0,NewP0_tl,NewP),
	$transform_arrow(PRest,L,NewP0_tl).

$transform_arrow1([],_,[]).
$transform_arrow1([L|LRest],P,[(P -> L)|LPRest]) :-
	$transform_arrow1(LRest,P,LPRest).

$transform3(_,_,_,[],[]).
$transform3(H,C0,C1,[Body|BRest],[rule(H,Body,C0,C1)|CRest]) :-
	$transform3(H,C0,C1,BRest,CRest).

$trans_check(pred(_,_,_,_,Cl)) :- $trans_check0(Cl).

$trans_check0([Cl|Crest]) :- $trans_check1(Cl) ; $trans_check(Crest).

$trans_check1(rule(H,B,_,_)) :- $trans_check2(B).

$trans_check2(','(C1,C2)) :- $trans_check2(C1) ; $trans_check2(C2).
$trans_check2('->'(C1,C2)) :- $trans_check2(C1) ; $trans_check2(C2).
$trans_check2(';'(_,_)).
$trans_check2(not(_)).

/*  given a list of predicate definitions, "$trans_get_deps" returns a list of
    terms "pred(P,N,Deps,StL,EqFlag)" where P/N is the predicate, Deps
    a list of pairs (P,N) giving the predicates it depends on, StL is
    a flag set to 0 if it is a straight-line program, 1 otherwise; and
    EqFlag is a flag set to 1 if any clause of the predicate contains
    explicit unifications via =/2, and to 0 otherwise.  		*/

$trans_get_deps([],[]).
$trans_get_deps([pred(P,N,_,_,Clauses)|PRest],
		[pred(P,N,Deps,StL,EqFlag)|DRest]) :-
	$trans_get_deps_2(Clauses,Deps,StL,EqFlag),
	$trans_get_deps(PRest,DRest).

$trans_get_deps_2([],[],0,0).
$trans_get_deps_2([fact(Fact,_)|ClRest],Deps,StL,EqFlag) :-
	$trans_get_deps_2(ClRest,Deps,StL,EqFlag).
$trans_get_deps_2([rule(H,B,_,_)|ClRest],Deps,StL,EqFlag) :-
	$trans_get_deps_3(B,Deps1,Stl0,EqFlag0),
	$trans_get_deps_2(ClRest,Deps2,Stl1,EqFlag1),
	$merge(Deps1,Deps2,Deps),
	$logical_or(Stl0,Stl1,StL),
	$logical_or(EqFlag0,EqFlag1,EqFlag).

$trans_get_deps_3(','(C1,C2),Deps,StL,EqFlag) :- !,
	$trans_get_deps_3(C1,Deps1,StL0,EqFlag0),
	$trans_get_deps_3(C2,Deps2,StL1,EqFlag1),
	$merge(Deps1,Deps2,Deps),
	$logical_or(StL0,StL1,StL),
	$logical_or(EqFlag0,EqFlag1,EqFlag).
$trans_get_deps_3('->'(C1,C2),Deps,StL,EqFlag) :- !,
	$trans_get_deps_3(C1,Deps1,StL0,EqFlag0),
	$trans_get_deps_3(C2,Deps2,StL1,EqFlag1),
	$merge(Deps1,Deps2,Deps),
	$logical_or(StL0,StL1,StL),
	$logical_or(EqFlag0,EqFlag1,EqFlag).
$trans_get_deps_3(';'(C1,C2),Deps,1,EqFlag) :- !,
	$trans_get_deps_3(C1,Deps1,_,EqFlag0),
	$trans_get_deps_3(C2,Deps2,_,EqFlag1),
	$merge(Deps1,Deps2,Deps),
	$logical_or(EqFlag0,EqFlag1,EqFlag).
$trans_get_deps_3(not(_),[],1,0) :- !.	    /* calling patterns unaffected */
$trans_get_deps_3(Call,[],0,1) :- $functor(Call,'=',2), !.
$trans_get_deps_3(Call,DepList,0,0) :-
	$functor(Call,P,N), 
	(($inline(P,N), DepList = []) ;
	 (P = '!', N = 0, DepList = []) ;
	 (DepList = [(P,N)])
	).

/*  given a list of (Pred,Arity) pairs, "$trans_get_dep_len" computes the
    total length of the success list tuples for these predicates.	*/

/*  given a dependency list, "$trans_get_affect_list" returns, for each
    predicate, a list of predicates which are affected by it.		*/

$trans_get_affect_list([],_,[]).
$trans_get_affect_list([pred(P,N,DList,StL,EqFlag)|Rest],BDList,
		[pred(P,N,AList,StL,EqFlag)|ARest]) :-
	$trans_get_affect_list_1(P,N,BDList,AList),
	$trans_get_affect_list(Rest,BDList,ARest).

$trans_get_affect_list_1(P,N,[],[]).
$trans_get_affect_list_1(P,N,[pred(P1,N1,DepList,_,_)|DRest],AList) :-
	(($member((P,N),DepList), !,
	  $trans_get_affpred(P1,N1,AffP),
	  AList = [AffP|ARest]
	 ) ;
	 AList = ARest
	),
	$trans_get_affect_list_1(P,N,DRest,ARest).

$trans_get_affpred(P,N,(AffP1,AffP0)) :-
	$name(P,Pname),$name(N,Nname),
	$append(Nname,Pname,N1name),
	$name(AffP,[36,97,102,102,95|N1name]),	/* "$aff_" prefix */
	$univ(AffP1,[AffP,1]), $univ(AffP0,[AffP,0]).

$trans_get_affs1([],[]).
$trans_get_affs1([(Aff1,Aff0)|ARest], [Aff1|ARest1]) :-
	$trans_get_affs1(ARest,ARest1).

/* ---------------------------- $transform1.P ---------------------------- */

